#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <fftwnd_mpi.h>

#define DRAND() ((((double) rand()) / (double)RAND_MAX) - .5)

#ifdef FFTW_ENABLE_FLOAT
#define TOLERANCE (1e-3)
#else
#define TOLERANCE (1e-8)
#endif

#define MAX_FACTOR 13

int rand_small_factors(int N)
{
     int f, n = 1;

     f = rand() % MAX_FACTOR + 1;

     while (n * f <= N) {
          n *= f;
          f = rand() % MAX_FACTOR + 1;
     }

     return n;
}

#define MAX_N 65536

void random_dims(int rank, int *n)
{
     int maxsize, dim;

     maxsize = pow(MAX_N, 1.0 / rank);
     if (maxsize < 1)
          maxsize = 1;

     for (dim = 0; dim < rank; ++dim)
          n[dim] = rand_small_factors(maxsize);
}

#define SQR(x) ((x)*(x))

void check_equal(FFTW_COMPLEX a, FFTW_COMPLEX b)
{
     FFTW_REAL absa, absb, absdiff;

     absa = sqrt(SQR(c_re(a)) + SQR(c_im(a)));
     absb = sqrt(SQR(c_re(b)) + SQR(c_im(b)));
     absdiff = sqrt(SQR(c_re(b) - c_re(a)) + SQR(c_im(b) - c_im(a)));
     
     if (absdiff > TOLERANCE * (absa + absb) * 0.5)
	  fftw_die("ERROR: incorrect value in result!\n");
}

#define MIN_RANK 2
#define MAX_RANK 5

void test_fftwnd_mpi(int my_pe, int rank, int *n, fftw_direction dir,
		     int verbose)
{
     fftwnd_mpi_plan p_mpi;
     fftwnd_plan p_uni;
     FFTW_COMPLEX *a_uni, *a_mpi = 0;
     int dim, N, n0, n1, nrest, i0, i1, irest;
     int local_nx, local_x_start, local_ny_after_transpose,
	  local_y_start_after_transpose, total_local_size;

     if (my_pe == 0) {
	  printf("Testing %d", n[0]);
	  for (dim = 1; dim < rank; ++dim)
	       printf("x%d",n[dim]);
	  printf(" %s transform...", 
		 dir == FFTW_FORWARD ? "forward" : "backward");
	  fflush(stdout);
     }

     if (rank < MIN_RANK) {
	  if (my_pe == 0) printf("illegal rank.\n");
	  return;
     }

     p_mpi = fftwnd_mpi_create_plan(MPI_COMM_WORLD,
				    rank,n,dir,FFTW_ESTIMATE|FFTW_IN_PLACE);
     p_uni = fftwnd_create_plan(rank,n,dir,FFTW_ESTIMATE|FFTW_IN_PLACE);


     if (my_pe == 0) { printf("..."); fflush(stdout); }

     N = 1;
     for (dim = 0; dim < rank; ++dim)
	  N *= n[dim];

     a_uni = fftw_malloc(sizeof(FFTW_COMPLEX) * N);

     fftwnd_mpi_local_sizes(p_mpi, &local_nx, &local_x_start,
			    &local_ny_after_transpose,
			    &local_y_start_after_transpose,
			    &total_local_size);

     if (total_local_size > 0)
       a_mpi = fftw_malloc(sizeof(FFTW_COMPLEX) * total_local_size);

     n0 = n[0];
     n1 = n[1];
     nrest = N / (n0 * n1);
     
     for (i0 = 0; i0 < n0; ++i0)
	  for (i1 = 0; i1 < n1; ++i1)
	       for (irest = 0; irest < nrest; ++irest) {
		    c_re(a_uni[(i0 * n1 + i1) * nrest + irest])
			 = DRAND();
		    c_im(a_uni[(i0 * n1 + i1) * nrest + irest])
			 = DRAND();
	       }

     for (i0 = 0; i0 < local_nx; ++i0)
	  for (i1 = 0; i1 < n1; ++i1)
	       for (irest = 0; irest < nrest; ++irest) {
		    c_re(a_mpi[(i1 * local_nx + i0) * nrest + irest])
			 = c_re(a_uni[((i0 + local_x_start) * n1 
				       + i1) * nrest + irest]);
		    c_im(a_mpi[(i1 * local_nx + i0) * nrest + irest])
			 = c_im(a_uni[((i0 + local_x_start) * n1 
				       + i1) * nrest + irest]);
	       }

     if (my_pe == 0) { printf("..."); fflush(stdout); }

     fftwnd(p_uni,1,a_uni,1,0,NULL,0,0);

     if (my_pe == 0) { printf("..."); fflush(stdout); }

     fftwnd_mpi(p_mpi,1,a_mpi,FFTW_NORMAL_ORDER);

     if (my_pe == 0) { printf("..."); fflush(stdout); }

     for (i0 = 0; i0 < local_nx; ++i0)
          for (i1 = 0; i1 < n1; ++i1)
               for (irest = 0; irest < nrest; ++irest)
                    check_equal(a_mpi[(i1 * local_nx + i0) * nrest + irest],
				a_uni[((i0 + local_x_start) * n1
                                       + i1) * nrest + irest]);

     if (my_pe == 0) { printf("..."); fflush(stdout); }

     if (a_mpi)
       fftw_free(a_mpi);
     fftw_free(a_uni);

     fftwnd_destroy_plan(p_uni);
     fftwnd_mpi_destroy_plan(p_mpi);

     fftw_check_memory_leaks();

     if (my_pe == 0)
	  printf("ok.\n");
}

int main(int argc, char **argv)
{
     int my_pe, i, rank = -1, n[MAX_RANK];
     int seed;

     MPI_Init(&argc,&argv);     
     MPI_Comm_rank(MPI_COMM_WORLD,&my_pe);

     seed = time(NULL);
     MPI_Bcast(&seed, 1, MPI_INT, 0, MPI_COMM_WORLD);
     srand(seed);

     fftw_die_hook = fftw_mpi_die; /* call MPI_Abort on failure */

     if (argc >= 2)
	  rank = atoi(argv[1]);

     if (rank >= MIN_RANK && argc >= 2 + rank) {
	  for (i = 0; i < rank; ++i) {
	       n[i] = atoi(argv[2 + i]);
	       if (n[i] <= 0)
	            fftw_die("Illegal param: dimensions must be positive.\n");
	  }
	  
	  test_fftwnd_mpi(my_pe,rank,n,FFTW_FORWARD,1);
	  test_fftwnd_mpi(my_pe,rank,n,FFTW_BACKWARD,1);
     }
     else {
	  int count = 0, rand_rank = rank < MIN_RANK;
	  while (1) {
	       if (count++ % 50 == 0 && my_pe == 0)
		    printf("Performing random tests (goes on forever).\n");
	       if (rand_rank)
	            rank = rand() % (MAX_RANK - MIN_RANK + 1) + MIN_RANK;
	       random_dims(rank,n);
	       test_fftwnd_mpi(my_pe, rank,n,
			       rand() % 2 ? FFTW_FORWARD : FFTW_BACKWARD, 0);
	  }
     }
     
     fftw_check_memory_leaks();

     MPI_Finalize();
     
     return 0;
}
